#!/usr/bin/env python
# -*- coding: utf-8 -*-

# wbfs2fat.py - WBFS to FAT32 conversion
# version 0.8.6 - 2012-07-27
# (c) 2012 by PsyBlade - all rights reserved
# http://gbatemp.net/t291320-wbfs2fat-py

if __name__ == "__main__":
    import wbfs2fat, sys
    wbfs2fat.main()
    sys.exit(0)

import cfg, sys
import string, struct, math, os, time, subprocess, ctypes, optparse
import re, io

sw2B=struct.Struct(">B")
sw2H=struct.Struct(">H")
sw2I=struct.Struct(">I")
sw2b=struct.Struct(">?")
sf2B=struct.Struct("<B")
sf2H=struct.Struct("<H")
sf2I=struct.Struct("<I")
unit=["B","KiB","MiB","GiB", "TiB"]


def findparts():
    if sys.platform in ["linux2","cygwin"]:
        partfile = open("/proc/partitions","rt")
        partfile.readline()
        partfile.readline()
        parts=[("/dev/"+line.split()[3], "/dev/"+line.split()[3]) for line in partfile]
        partfile.close()
    elif sys.platform == "win32":
        import wmi
        w32 = wmi.WMI ()
        parts=[]
        for disk in w32.Win32_LogicalDisk():
            parts.append(("\\\\.\\"+disk.DeviceID,disk.DeviceID))
    return parts


def scanwbfs(device,name=None):
    if not name:
        name = device
    try:
        dev = io.BufferedReader(io.FileIO(device,"rb"))
        wbfs = Wbfs(dev.read(12))
        dev.seek(0)
        wbfs.raw = dev.read(1<<wbfs.block_size_p)
        dev.close()
    except IOError:
        if sys.platform == "win32" and name==device:
            return scanwbfs("\\\\.\\"+device, name)
        else:
            return(name, name, None, None, "permission denied", 7)
    except WbfsError as e:
        return(device, name, None, None, e.args[1], e.args[0])
    msg = "OK"
    status = 0
    try:
        wbfs.analyse()
    except WbfsError as e:
        msg = e.args[1]
        if not cfg.options.ignore6 and e.args[0] == 6:
            status = e.args[0]
    size = wbfs.sect_cnt << wbfs.sect_size_p
    n=0
    while size>>14:
        size = size>>10
        n+=1
    return(device, name, str(size)+" "+unit[n], len(wbfs.games), msg ,status)


def rename(old,new):
    cfg.report(2,"\t"+old+"->"+new)
    try:
        os.renames(old,new)
    except Exception as e:
        cfg.report(0,"\nERROR: rename to "+new+" failed: "+str(e.args[0])+" "+str(e.args[1]))
        return 0
    return 1


class WbfsError(Exception):
    pass


class Wbfs:
    def __init__(self,head):
        self.head=head
        self.sect_cnt, self.sect_size_p, self.block_size_p, self.version = struct.unpack_from(">IBBBx", head,4)
        if head[0:4] != "WBFS":
            raise WbfsError(3,"Not a WBFS partition (missing magic)")
        if self.version > 1:
            raise WbfsError(4,"Unknown WBFS Version:"+str(self.version))

    def block2adr(self,block):
        return block*self.block_size

    def analyse(self):
        self.sect_size = 1<<self.sect_size_p
        self.block_size = 1<<self.block_size_p
        self.block_cnt = self.sect_cnt>>self.block_size_p-self.sect_size_p
        self.block_sects = 1<<self.block_size_p-self.sect_size_p

        self.clust_size_p = sorted([self.sect_size_p, math.frexp((self.sect_cnt<<self.sect_size_p)/268435444)[1], 15, math.frexp((self.sect_cnt<<self.sect_size_p)/65525)[1]-1, self.sect_size_p+7])[2]
        self.clust_size = 1<<self.clust_size_p
        self.clust_sects_p = self.clust_size_p - self.sect_size_p
        self.clust_sects = 1<<self.clust_sects_p
        self.fat_head_sects = Fat.res_sects + (self.sect_cnt-1>>self.clust_sects_p+self.sect_size_p-2)+1
        self.fat_blocks = (self.fat_head_sects-1>>self.block_size_p-self.sect_size_p)+1
        self.block_clusts = self.block_sects>>self.clust_sects_p

        self.file_blocks_max = (1<<32)-1>>self.block_size_p
        self.disc_blocks = 143432*2 >> ( self.block_size_p - 15 )
        self.slot_size = ((0x0100 + 2*self.disc_blocks)-1>>self.sect_size_p)+1<<self.sect_size_p

        back=open(cfg.options.back_dir+"/wbfs2fat.debug","wb")
        back.write(self.raw)
        back.close()

        files = 0
        used = []
        self.games = {}
        for slot in range(self.sect_size-12):
            if sw2b.unpack_from(self.raw,12+slot)[0]:
                block_map = [(-1,0)]
                head = self.raw[self.sect_size+slot*self.slot_size:self.sect_size+slot*self.slot_size+0x0080]
                game = str(head[0:6]).strip()
                cfg.report (3, "\t"+game)
                for src in range(self.disc_blocks):
                    dest = sw2H.unpack_from(self.raw, self.sect_size+0x0100+slot*self.slot_size+src*2)[0]
                    if dest:
                        block_map.append((src,dest))
                        cfg.report (4,str((src,dest)))
                        used.append((dest, game))
                self.games[game] = (block_map,head)
                files += (len(block_map)-1)/self.file_blocks_max+1
        self.fat_blocks += files >> self.block_size_p-10

        used.sort()
        rel=[]
        free=[]
        last=0
        lastgame=""
        error6=""
        error8=""
        for (block, game) in used:
            if block >= self.block_cnt:
                error7="block "+str(block)+" claimed by "+str(game)+" but only "+str(self.block_cnt)+" blocks in filesystem\n"
                cfg.report(1,"\n"+error7)
            elif block == last:
                error6="block "+str(block)+" claimed by "+str(lastgame)+" and "+str(game)+"\n"
                cfg.report(1,"\n"+error6)
                rel.append((block, game))
            elif block < self.fat_blocks: rel.append((block, game))
            elif block > max(last+1,self.fat_blocks): free.extend(range(max(last+1,self.fat_blocks),block))
            last = block
            lastgame = game
        free.extend(range(last+1,self.block_cnt))
        self.remap = dict(zip(rel,free))
        for (_,(blocks,_)),new in zip(self.games.iteritems(),free[len(rel):]):
            blocks[0]=(-1,new)

        n_blocks = len(rel)+len(self.games)-len(free)
        if n_blocks > 0:
            raise WbfsError(5,str(n_blocks)+" aditional free Blocks of "+str(1<<self.block_size_p%10)+" "+unit[self.block_size_p/10]+" needed")
        if error6:
            raise WbfsError(6,'WBFS is corrupted: Blocks used multiple Times (try "wwt help verify" to fix or use ignore-error-6 to ignore)')
        if error8:
            raise WbfsError(8,'WBFS is corrupted: Blocks used are outside filesystem boundry (try deleting them to fix)')



class Fat:
    res_sects_need = 7
    res_sects_flex = 64
    res_sects=res_sects_need + res_sects_flex

    def __init__(self,wbfs):
        self.wbfs = wbfs
        self.raw = ctypes.create_string_buffer(wbfs.fat_blocks * wbfs.block_size)
        self.res_sects = (wbfs.block_sects - wbfs.fat_head_sects + Fat.res_sects_flex) % wbfs.clust_sects + Fat.res_sects_need
        self.fat_head_sects = wbfs.fat_head_sects - Fat.res_sects + self.res_sects

    def block2clust(self,block):
        return self.adr2clust(self.wbfs.block2adr(block))

    def adr2clust(self,adr):
        return (adr-self.clust_start)//(self.wbfs.clust_size)

    def clust2adr(self,clust):
        return self.clust_start + clust*self.wbfs.clust_sects*self.wbfs.sect_size

    def clust2fat(self,clust):
        return self.fat_start + (clust<<2)

    def mkfs(self):
        cfg.report(1,"\nW2F: creating FAT filesystem")
        wbfs = self.wbfs
        fat = self.raw
        self.fat_start = wbfs.sect_size * self.res_sects
        self.clust_start = wbfs.sect_size * (self.fat_head_sects - 2*wbfs.clust_sects)

        #Sector 0
        #0x00  3  Jump
        struct.pack_into("<BBB",fat,0x00,0x0eb,0x058,0x090)
        #0x03  8  OEM Name
        fat[0x03:0x0b] = 'wbfs2fat'
        #0x0b  2  Bytes per sector
        sf2H.pack_into(fat,0x0b,wbfs.sect_size)
        #0x0d  1  Sectors per cluster
        sf2B.pack_into(fat,0x0d,wbfs.clust_sects)
        #0x0e  2  Reserved sector count
        sf2H.pack_into(fat,0x0e,self.res_sects)
        #0x10  1  Number of FATs
        sf2B.pack_into(fat,0x10,1)
        #0x11  2  0
        #0x13  2  Total sectors 2
        if wbfs.sect_cnt <= 65535:
            sf2H.pack_into(fat,0x13,wbfs.sect_cnt)
        #0x15  1  Media descriptor
        sf2B.pack_into(fat,0x15,0x0f8)
        #0x16  2  0
        #0x18  2  Sectors per track
        #0x1a  2  Number of heads
        #0x1c  4  Count of hidden sectors preceding
        #0x20  4  Total sectors 4
        if wbfs.sect_cnt > 65535:
            sf2I.pack_into(fat,0x20,wbfs.sect_cnt)
        #0x24  4  Sectors per FAT
        sf2I.pack_into(fat,0x24,self.fat_head_sects - self.res_sects)
        #0x28  2  FAT Flags
        #0x2a  2  Version (Defined as 0)
        #0x2c  4  Cluster number of root directory start
        sf2I.pack_into(fat,0x2c,2)
        #0x30  2  Sector number of FS Information Sector
        sf2H.pack_into(fat,0x30,1)
        #0x32  2  Sector number of a copy of this boot sector (0 if no backup copy exists)
        sf2H.pack_into(fat,0x32,6)
        #0x34  12 Reserved
        #0x40  1  Physical Drive Number
        sf2B.pack_into(fat,0x40,0x080)
        #0x41  1  Reserved
        #0x42  1  Extended boot signature
        sf2B.pack_into(fat,0x42,0x029)
        #0x43  4  ID (serial number)
        #0x47  11  Volume Label
        fat[0x47:0x52] = 'wbfs2fat   '
        #0x52  8  FAT file system type: "FAT32   "
        fat[0x52:0x5a] = 'FAT32   '
        #0x5a  420  Operating system boot code
        #0x1FE  2  Boot sector signature (0x55 0xAA)
        sf2H.pack_into(fat,0x01FE,0x0AA55)

        #sector 1
        #0x00  4  FS information sector signature
        sf2I.pack_into(fat,wbfs.sect_size+0x00,0x041615252)
        #0x04  480  Reserved (byte values are 0x00)
        #0x1e4  4  FS information sector signature
        sf2I.pack_into(fat,wbfs.sect_size+0x01e4,0x061417272)
        #0x1e8  4  Number of free clusters on the drive, or -1 if unknown
        struct.pack_into("<i",fat,wbfs.sect_size+0x01e8,-1)
        #0x1ec  4  Number of the most recently allocated cluster
        #0x1f0  14  Reserved (byte values are 0x00)
        #0x1fe  2  FS information sector signature
        sf2H.pack_into(fat,wbfs.sect_size+0x01fe,0x0AA55)

        #sector 6
        #backup of sector 0
        fat[6*wbfs.sect_size:7*wbfs.sect_size]=fat[0:wbfs.sect_size]

        #FAT
        sf2I.pack_into(fat,self.fat_start+0x00,0x0fffffff8)
        sf2I.pack_into(fat,self.fat_start+0x04,0x00ffffff8)
        sf2I.pack_into(fat,self.fat_start+0x08,0x00ffffff8)


    def mkfiles(self):
        cfg.report(1,"\nW2F: creating placeholder Files")
        self.files={}
        clust = 2
        adr = self.clust2adr(clust)
        ltime = time.localtime()
        fat_date = sf2H.pack(((ltime[0]-1980)<<9) + (ltime[1]<<5) + ltime[2])
        fat_time = sf2H.pack((ltime[3]<<11) + (ltime[4]<<5) + (ltime[5]>>1))
        self.raw[adr:adr+0x0b]='wbfs2fat   '
        sf2B.pack_into(self.raw,adr+0x0b,0x08)
        self.raw[adr+0x0E:adr+0x10] = fat_time
        self.raw[adr+0x10:adr+0x12] = fat_date
        self.raw[adr+0x12:adr+0x14] = fat_date
        self.raw[adr+0x16:adr+0x18] = fat_time
        self.raw[adr+0x18:adr+0x1a] = fat_date
        adr+=0x020
        for id6,(blocks,_) in self.wbfs.games.iteritems():
            for ext in range(len(blocks)/self.wbfs.file_blocks_max+1):
                self.files[(id6,ext)]=adr
                #0x00  8  name
                self.raw[adr:adr+0x0b]='           '
                self.raw[adr:adr+min(len(id6),8)]=id6[:6]
                #0x08  3  extension
                self.raw[adr+0x08:adr+0x08+min(len(str(ext)),3)]=str(ext)[:3]
                #0x0b  1  attributes
                #0x0c  1  reserved
                #0x0d  1  ctime fine
                #0x0E  2  ctime time
                self.raw[adr+0x0E:adr+0x10] = fat_time
                #0x10  2  ctime date
                self.raw[adr+0x10:adr+0x12] = fat_date
                #0x12  2  atime date
                self.raw[adr+0x12:adr+0x14] = fat_date
                #0x14  2  cluster (high)
                #0x16  2  mtime time
                self.raw[adr+0x16:adr+0x18] = fat_time
                #0x18  2  mtime date
                self.raw[adr+0x18:adr+0x1a] = fat_date
                #0x1a  2  cluster (low)
                #0x1c  4  size
                adr+=0x020
                if adr >= self.clust2adr(clust+1):
                    struct.pack_into("<II",self.raw,self.clust2fat(clust),clust+1,0x0ffffff8)
                    clust = clust+1
                    adr = self.clust2adr(clust)


    def link(self):
        cfg.report(1,"\nW2F: linking used Blocks to Files")
        adr=0
        dir_adr=0
        fat=self.raw
        for id6,(blocks,_) in self.wbfs.games.iteritems():
            cfg.report(2,"\t"+str(id6))
            ext=0
            file_blocks=0
            for (_,dest) in blocks:
                if (dest, id6) in self.wbfs.remap: dest=self.wbfs.remap[dest, id6]
                c_block_clust = self.block2clust(dest)
                if file_blocks>=self.wbfs.file_blocks_max:
                    sf2I.pack_into(fat,adr,0x0FFFFFFFF)
                    sf2I.pack_into(fat,dir_adr+0x1c,self.wbfs.block_size*file_blocks)
                    ext+=1
                    file_blocks=0
                if file_blocks==0:
                    cfg.report(2," "+str(ext))
                    dir_adr=self.files[(id6,ext)]
                    sf2H.pack_into(fat,dir_adr+0x014,c_block_clust>>16)
                    sf2H.pack_into(fat,dir_adr+0x01a,c_block_clust%2**16)
                    adr=self.clust2fat(c_block_clust)
                for c_clust in range(c_block_clust,c_block_clust+self.wbfs.block_clusts):
                    sf2I.pack_into(fat,adr,c_clust)
                    adr=self.clust2fat(c_clust)
                file_blocks+=1
            sf2I.pack_into(fat,adr,0x0FFFFFFFF)
            sf2I.pack_into(fat,dir_adr+0x1c,self.wbfs.block_size*file_blocks)



class Device:
    def __init__(self, name):
        self.dev_name=name

    def open_ro(self):
        self.dev = io.BufferedReader(io.FileIO(self.dev_name, "rb"))

    def open_rw(self):
        self.dev = io.BufferedRandom(io.FileIO(self.dev_name, "r+b"))

    def close(self):
        self.dev.close()

    def init_wbfs(self):
        cfg.report(1,"\nW2F: checking for WBFS")
        try:
            self.dev.seek(0)
            self.wbfs = Wbfs(self.dev.read(12))
        except IOError:
            print "ERROR 7 : permission denied"
            sys.exit(7)
        return self.wbfs

    def read_wbfs(self):
        cfg.report (1,"\nW2F: analysing WBFS")
        self.dev.seek(0)
        self.wbfs.raw=self.dev.read(1<<self.wbfs.block_size_p)

    def backup(self):
        cfg.report(1,"\nW2F: creating Backup of overlap")
        bname=cfg.options.back_dir+"/wbfs."+re.sub(":","",self.dev_name.strip().split('\\')[-1].split('/')[-1]+"."+str(int(time.time())))+".back"
        back=open(bname,"wb")
        self.dev.seek(0)
        for _ in range(self.wbfs.fat_blocks):
            back.write(self.dev.read(self.wbfs.block_size))
        back.close()

    def restore(self, fname):
        cfg.report(1,"\nW2F: restoring Backup")
        self.open_rw()
        file = open(fname, "rb")
        block = file.read(1<<20)
        while len(block)>0:
            self.dev.write(block)
            block = file.read(1<<20)
        file.close()
        self.close()
        cfg.report(1,"\nW2F: Restore done")

    def init_fat(self):
        self.fat=Fat(self.wbfs)
        return self.fat

    def write_head(self):
        cfg.report(1,"\nW2F: creating WBFS File Headers")
        wbfs = self.wbfs
        dev = self.dev
        sect_size = wbfs.sect_size
        for _,(blocks,head) in wbfs.games.iteritems():
            _,new = blocks[0]
            buff=ctypes.create_string_buffer(wbfs.block_size)
            buff[0:12]=wbfs.head
            sw2I.pack_into(buff,4,len(blocks)*wbfs.block_sects)
            sw2B.pack_into(buff,12,0x01)
            buff[sect_size:sect_size+len(head)]=head
            cnt = 0
            for src, _ in blocks[1:]:
                cnt+=1
                sw2H.pack_into(buff,sect_size+0x0100+src*2,cnt)
            freesize = (cnt-1 >> 3+wbfs.sect_size_p) + 1 << wbfs.sect_size_p
            for adr in range(wbfs.block_size-2*freesize, wbfs.block_size-freesize, 4):
                if cnt < 0:
                    sw2I.pack_into(buff,adr,0xffffffff)
                elif cnt < 32:
                    sw2I.pack_into(buff,adr,0xffffffff - (1<<cnt)+1)
                cnt-=32
            dev.seek(wbfs.block2adr(new))
            dev.write(buff)

    def write_relocate(self):
        cfg.report(1,"\nW2F: relocating overlaping Blocks")
        dev = self.dev
        block2adr = self.wbfs.block2adr
        for (old,  _), new in sorted(self.wbfs.remap.iteritems()):
            cfg.report(2,"\t"+str(old)+"->"+str(new))
            dev.seek(block2adr(old))
            data = dev.read(self.wbfs.block_size)
            dev.seek(block2adr(new))
            dev.write(data)

    def write_fat(self):
        cfg.report(1,"\nW2F: writing FAT to Disk")
        self.dev.seek(0)
        for i in range(self.wbfs.fat_blocks):
            self.dev.write(self.fat.raw[i*self.wbfs.block_size:(i+1)*self.wbfs.block_size])

    def fsck(self):
        if sys.platform == "linux2":
            cfg.report(1,"\nW2F: checking FAT for Errors")
            fsck=['fsck.vfat', '-a', self.dev_name]
            if cfg.options.verbosity>1: fsck.append("-v")
            subprocess.call(fsck)

    def ntfs(self):
        if sys.platform == "win32":
            cfg.report(1,"\nW2F: converting to NTFS")
            if subprocess.call(['convert', self.dev_name, "/x", "/nosecurity", "/fs:ntfs"]): sys.exit()

    def mount(self):
        if sys.platform == "linux2":
            cfg.report(1,"\nW2F: mounting FAT")
            if subprocess.call(['mount', '-o', 'loop,umask=000', self.dev_name, cfg.options.fat_dir]): sys.exit()

    def move(self):
        cfg.report(1,"\nW2F: moving Files to destination")
        for (id6,ext) in self.fat.files.keys():
            old = cfg.options.fat_dir+"/"+id6+"."+str(ext)
            if ext == 0: ext="s"
            fname = id6+".wbf"+str(ext)
            dname = str(self.wbfs.games[id6][1][0x20:]).strip(' \t\r\n\0')+" ["+id6+"]"
            dname = dname.translate(string.maketrans('\/|*?^"<>',":::~  '()"))
            dname = re.sub(':',' - ',dname)
            dname = re.sub(' +',' ',dname)
            if not rename(old,cfg.options.fat_dir+"/wbfs/"+dname+"/"+fname):
                if not rename(old,cfg.options.fat_dir+"/wbfs/"+fname):
                    cfg.report(0,"\nERROR: please manually rename "+old)

    def convert(self):
        self.open_ro()
        try:
            wbfs = self.init_wbfs()
        except WbfsError as e:
            if cfg.options.mode == "convert":
                print "ERROR",e.args[0],":",e.args[1]
                sys.exit(e.args[0])
            else:
                raise
        self.read_wbfs()
        self.close()

        try:
            wbfs.analyse()
        except WbfsError as e:
            if not cfg.options.ignore6 and e.args[0] == 6:
                if cfg.options.mode == "convert":
                    print "ERROR",e.args[0],":",e.args[1]
                    sys.exit(e.args[0])
                else:
                    raise

        fat = self.init_fat()
        fat.mkfs()
        fat.mkfiles()
        fat.link()

        self.open_rw()
        if cfg.options.backup: self.backup()
        cfg.confirm("W2F: ready for conversion")
        cfg.report(1,"\nWARNING: start of dangerous part")
        self.write_head()
        if wbfs.remap: self.write_relocate()
        self.write_fat()
        self.close()
        cfg.report(1,"\nW2F: end of dangerous part")

        if cfg.options.fsck: self.fsck()
#        self.ntfs()
        if cfg.options.rename:
            self.mount()
            self.move()

        cfg.report(1,"\nW2F: Conversion completed\n")


def main():
    usage = "usage: %prog [options] device"
    parser = optparse.OptionParser(usage,version = "%prog 0.8.6")
    parser.add_option("-m", "--mode", action="store", dest="mode", type="choice", choices=("scan","convert", "gui",),
                      help="'scan' for partitions, 'convert' a partition or start the 'gui'")
    parser.add_option("-b", "--backup_dir", dest="back_dir", help="directory to store backup to", default=".")
    parser.add_option("-v", "--verbose", action="count", dest="verbosity", default=1)
    parser.add_option("-q", "--quiet", action="store_const", const=0, dest="verbosity")
    Fgroup = optparse.OptionGroup(parser, "dangerous Options")
    Fgroup.add_option("-Y", "--yes", action="store_false", dest="confirm", default=1, help="asume yes on all questions")
    Fgroup.add_option("-X", "--no-backup", action="store_false", dest="backup", default=1)
    Fgroup.add_option("-R", "--no-rename", action="store_false", dest="rename", default=1)
    Fgroup.add_option("--ignore-error-6", action="store_true", dest="ignore6", default=0,  help="copy blocks instead of abort on error 6")
    Lgroup = optparse.OptionGroup(parser, "special Linux Options")
    Lgroup.add_option("-d", "--mount-point", dest="fat_dir", metavar="DIR", help="directory to mount device to (for rename)")
    Lgroup.add_option("-F", "--no-fsck", action="store_false", dest="fsck", default=1)
    Cgroup = optparse.OptionGroup(parser, "special Cygwin Options")
    Cgroup.add_option("-l", "--drive-letter", dest="fat_dir", metavar="X", help="drive letter of the WBFS partition (for rename)")
    Bgroup = optparse.OptionGroup(parser, "special Backup Options")
    Bgroup.add_option("--restore", dest="restore", metavar="BACKUP", help="restore a Backup after a failed conversion")
    parser.add_option_group(Fgroup)
    parser.add_option_group(Lgroup)
    parser.add_option_group(Cgroup)
    parser.add_option_group(Bgroup)
    (cfg.options, args) = parser.parse_args()

    if len(args) == 1:
        cfg.dev_name = args[0]

    if sys.platform == "linux2":
        if not cfg.options.fat_dir:
            cfg.options.fat_dir = "./tmp"

    elif sys.platform == "cygwin":
        if cfg.options.fat_dir:
            cfg.options.fat_dir="/cygdrive/"+cfg.options.fat_dir.lower().split(":")[0][-1]

    elif sys.platform == "win32":
        if cfg.dev_name:
            cfg.options.fat_dir=string.lstrip(cfg.dev_name, "\.")[0]+":"
            cfg.dev_name="\\\\.\\"+cfg.options.fat_dir

    if cfg.options.restore:
        if cfg.dev_name:
            dev = Device(cfg.dev_name)
            dev.restore(cfg.options.restore)
        else:
            print usage
            sys.exit(2)

    elif cfg.options.mode == "gui":
        import gui.gui
        gui.gui.startgui()

    elif cfg.options.mode == "scan":
        import cli
        if cfg.dev_name:
            cli.printwbfs([cfg.dev_name])
        else:
            for part in findparts():
                cli.printwbfs(part)

    elif cfg.options.mode == "convert":
        if cfg.dev_name:
            import cli
            dev = Device(cfg.dev_name)
            cli.checkconfig()
            dev.convert()
        else:
            print usage
            sys.exit(2)

    else:
        import gui.gui
        gui.gui.startgui()

    sys.exit(0)
